@@ -1,37 +1,34 @@
-import os
-import sys
 import webbrowser
 from pathlib import Path
 
+import sys
 import argparse
 
-from git import Repo
-from git.exc import InvalidGitRepositoryError
-
-import gitgud
-from gitgud import operations
+from gitgud import __version__, InitializationError
+from gitgud.operations import Operator, get_operator
 from gitgud.skills import all_skills
-from gitgud.skills.user_messages import all_levels_complete
-from gitgud.skills.user_messages import show_tree
-from gitgud.skills.user_messages import handle_solution_confirmation
-from gitgud.skills.user_messages import mock_simulate
-from gitgud.skills.user_messages import print_info
-from gitgud.skills.user_messages import show_skill_tree
-from gitgud.hooks import all_hooks
 
-
-class InitializationError(Exception):
-    pass
+from gitgud.user_messages import force_initializing
+from gitgud.user_messages import cant_init_repo_not_empty
+from gitgud.user_messages import deleting_and_initializing
+from gitgud.user_messages import all_levels_complete
+from gitgud.user_messages import show_tree
+from gitgud.user_messages import handle_load_confirm
+from gitgud.user_messages import rerun_with_confirm_for_solution
+from gitgud.user_messages import show_skill_tree
+from gitgud.user_messages.stateful import repo_already_initialized
 
 
 class GitGud:
     def __init__(self):
         # Only gets operator if Git Gud has been initialized
-        self.parser = argparse.ArgumentParser(prog='git gud')
+        self.parser = argparse.ArgumentParser(
+            prog='git gud',
+            description='A game to teach Git!')
         self.parser.add_argument(
                 '--version',
                 action='version',
-                version='%(prog)s ' + gitgud.__version__)
+                version='%(prog)s ' + __version__)
 
         load_description = '\n'.join([
             "Load a specific skill or level. This command can be used in several ways.",  # noqa: E501
@@ -84,14 +81,18 @@ class GitGud:
                 dest='command'
         )
 
+        self.subparsers.add_parser(
+                'debug',
+                description='Debug git-gud with necessary imports set up.'
+        )
         self.subparsers.add_parser(
                 'status',
                 help='Print out the name of the current level',
                 description='Print out the name of the current level')
         self.subparsers.add_parser(
-                'instructions',
-                help='Show the instructions for the current level',
-                description='Show the instructions for the current level')
+                'explain',
+                help='Show the explain for the current level',
+                description='Show the explain for the current level')
         self.subparsers.add_parser(
                 'goal',
                 help='Concisely show what needs to be done to complete the level.',  # noqa: E501
@@ -149,6 +150,7 @@ class GitGud:
                 description='Initialize the direcotry with a git repository and load the first level of Git Gud.'  # noqa: E501
         )
         init_parser.add_argument('--force', action='store_true')
+        init_parser.add_argument('--prettyplease', action='store_true')
 
         load_parser = self.subparsers.add_parser(
                 'load',
@@ -164,6 +166,7 @@ class GitGud:
                 metavar='level',
                 nargs='?',
                 help='Level to load')
+        load_parser.add_argument('--force', action="store_true")
 
         skills_parser = self.subparsers.add_parser(
                 'skills',
@@ -175,9 +178,22 @@ class GitGud:
                 'levels',
                 help='List levels in a skill',
                 description='List the levels in the specified skill or in the current skill if Git Gud has been initialized and no skill is provided. To see levels in all skills, use `git gud levels --all`.')  # noqa: E501
-        levels_parser.add_argument('skill_name', metavar='skill', nargs='?')
-        levels_parser.add_argument('--all', dest='opt_all', action='store_true', help="Prints all available skills with levels.")  # noqa: E501
-        levels_parser.add_argument('--short', dest='opt_short', action='store_true', help="Prints with the short name of skills/levels usable with `git gud load`.")  # noqa: E501
+
+        levels_parser.add_argument(
+                'skill_name',
+                metavar='skill',
+                nargs='?')
+        levels_parser.add_argument(
+                '-a',
+                '--all',
+                dest='opt_all',
+                action='store_true',
+                help="Prints all available skills with levels.")
+        levels_parser.add_argument(
+                '--short',
+                dest='opt_short',
+                action='store_true',
+                help="Prints with the short name of skills/levels usable with `git gud load`.")  # noqa: E501
 
         commit_parser = self.subparsers.add_parser(
                 'commit',
@@ -199,7 +215,7 @@ class GitGud:
             'help': self.handle_help,
             'init': self.handle_init,
             'status': self.handle_status,
-            'instructions': self.handle_instructions,
+            'explain': self.handle_explain,
             'goal': self.handle_goal,
             'reset': self.handle_reset,
             'reload': self.handle_reset,
@@ -213,43 +229,45 @@ class GitGud:
             'show': self.handle_show,
             'contributors': self.handle_contrib,
             'issues': self.handle_issues,
-            'solution': self.handle_solution
+            'solution': self.handle_solution,
+            'debug': self.handle_debug
         }
 
+        self.aliases = {
+            'h': 'help',
+            'i': 'issues',
+            's': 'status',
+            'e': 'explain',
+            'g': 'goal',
+            'r': 'reset',
+            't': 'test',
+            'l': 'load',
+            'c': 'commit',
+        }
+
+        self.parser.epilog = "The following aliases exist: "
+        for alias in self.aliases:
+            self.parser.epilog += "'{}' -> '{}', " \
+                    .format(alias, self.aliases[alias])
+        self.parser.epilog = self.parser.epilog[:-2]
+
+        for alias in self.aliases:
+            assert self.aliases[alias] in self.command_dict
+
     def is_initialized(self):
-        return operations.get_operator() is not None
+        return get_operator() is not None
 
-    def assert_initialized(self, skip_level_check=False):
+    def assert_initialized(self):
         if not self.is_initialized():
-            raise InitializationError('Git gud has not been initialized. Use "git gud init" to initialize')  # noqa: E501
-
-        if not skip_level_check:
-            try:
-                self.get_level()
-            except KeyError:
-                level_name = operations.get_operator().read_level_file()
-                raise InitializationError(
-                        'Currently loaded level does not exist: "{}"'
-                        .format(level_name))
+            raise InitializationError('Git Gud has not been initialized. Use "git gud init" to initialize.')  # noqa: E501
 
     def load_level(self, level):
-        # Clear remotes
-        self.assert_initialized(skip_level_check=True)
-        file_operator = operations.get_operator()
-        level_repo = file_operator.setup_repo()
-        if level_repo:
-            for remote in level_repo.remotes:
-                level_repo.delete_remote(remote)
-        else:
-            file_operator.setup_repo()
+        self.assert_initialized()
+        file_operator = get_operator()
         file_operator.clear_tracked_commits()
         level.setup()
         file_operator.write_level(level)
 
-    def get_level(self):
-        skill_name, level_name = operations.get_operator().read_level_file().split()  # noqa: E501
-        return all_skills[skill_name][level_name]
-
     def handle_help(self, args):
         if args.command_name is None:
             self.parser.print_help()
@@ -264,123 +282,64 @@ class GitGud:
 
     def handle_init(self, args):
         # Make sure it's safe to initialize
-        file_operator = operations.get_operator()
-        if not args.force:
-            # We aren't forcing
-            if file_operator and Path(file_operator.path) in Path(os.getcwd()).parents:  # noqa: E501:
-                print('Repo {} already initialized for git gud.'
-                      .format(file_operator.path))
-                print('Use --force to initialize {}.'.format(os.getcwd()))
+
+        file_operator = get_operator()
+        if file_operator:
+            if not args.force:
+                repo_already_initialized()
                 return
-
-            file_operator = operations.get_operator(os.getcwd())
-
-            if os.path.exists(file_operator.gg_path):
-                # Current directory is a git repo
-                print('Git gud has already initialized. Use --force to force initialize again.')  # noqa: E501
-                return
-            if os.path.exists(os.path.join(file_operator.git_path, 'HEAD')):
-                # Current directory is a git repo
-                print('Currently in a git repo. Use --force to force initialize here.')  # noqa: E501
-                return
-            elif len(os.listdir(file_operator.path)) != 0:
-                print('Current directory is nonempty. Use --force to force initialize here.')  # noqa: E501
-                return
-        else:
-            print('Force initializing Git Gud.')
-            if not file_operator:
-                file_operator = operations.get_operator(os.getcwd())
-
-        assert file_operator is not None
-        # After here, we initialize everything
-        try:
-            file_operator.repo = Repo(file_operator.path)
-        except InvalidGitRepositoryError:
-            file_operator.repo = Repo.init(file_operator.path)
-
-        # Disable pager so "git gud status" can use the output easily
-        file_operator.shutoff_pager()
-
-        if not os.path.exists(file_operator.gg_path):
-            os.mkdir(file_operator.gg_path)
-
-        # Git uses unix-like path separators
-        python_exec = sys.executable.replace('\\', '/')
-
-        for git_hook_name, module_hook_name, accepts_args in all_hooks:
-            path = os.path.join(file_operator.hooks_path, git_hook_name)
-            if accepts_args:
-                forward_stdin = 'cat - |'
-                passargs = ' "$@"'
             else:
-                forward_stdin = ''
-                passargs = ''
+                force_initializing()
+        elif len(list(Path.cwd().iterdir())) != 0:
+            if not (args.force and args.prettyplease):
+                cant_init_repo_not_empty()
+                return
+            else:
+                deleting_and_initializing()
 
-            with open(path, 'w+') as hook_file:
-                hook_file.write('#!/bin/bash' + os.linesep)
-                hook_file.write(
-                        forward_stdin +
-                        python_exec + ' -m gitgud.hooks.' + module_hook_name +
-                        passargs + os.linesep)
-                hook_file.write(
-                    "if [[ $? -ne 0 ]]" + os.linesep + ""
-                    "then" + os.linesep + ""
-                    "\t exit 1" + os.linesep + ""
-                    "fi" + os.linesep)
-
-            # Make the files executable
-            mode = os.stat(path).st_mode
-            mode |= (mode & 0o444) >> 2
-            os.chmod(path, mode)
+        file_operator = Operator(Path.cwd())
+        file_operator.init_gg()
 
         self.load_level(all_skills["0"]["1"])
 
     def handle_status(self, args):
-        if self.is_initialized():
-            try:
-                level = self.get_level()
-                level.status()
-            except KeyError:
-                level_name = operations.get_operator().read_level_file()
-                print('Currently on unregistered level: "{}"'
-                      .format(level_name))
-        else:
-            print("Git Gud not initialized.")
-            print('Initialize using "git gud init"')
-
-    def handle_instructions(self, args):
         self.assert_initialized()
-        self.get_level().instructions()
+        get_operator().get_level().status()
+
+    def handle_explain(self, args):
+        self.assert_initialized()
+        get_operator().get_level().explain()
 
     def handle_goal(self, args):
         self.assert_initialized()
-        self.get_level().goal()
+        get_operator().get_level().goal()
 
     def handle_reset(self, args):
         self.assert_initialized()
-
-        level = self.get_level()
+        level = get_operator().get_level()
         self.load_level(level)
 
     def handle_test(self, args):
         self.assert_initialized()
-        level = self.get_level()
+        level = get_operator().get_level()
         level.test()
 
     def handle_solution(self, args):
         self.assert_initialized()
-        current_level = self.get_level()
+        current_level = get_operator().get_level()
         if not args.confirm and \
                 not current_level.has_ever_been_completed():
-            handle_solution_confirmation(current_level)
+            rerun_with_confirm_for_solution(current_level)
         else:
             current_level.solution()
 
     def handle_level(self, args):
-        level = self.get_level()
+        self.assert_initialized()
+        level = get_operator().get_level()
         skill = level.skill
         show_skill_tree(
                 [skill, level],
+                True,
                 expand_skills=False)
 
     def handle_skills(self, args):
@@ -388,6 +347,7 @@ class GitGud:
         print()
         show_skill_tree(
             [skill for skill in all_skills],
+            bool(get_operator()),
             expand_skills=False,
             show_human_names=not args.opt_short
         )
@@ -407,7 +367,7 @@ class GitGud:
                     print('You may run "git gud levels --all" or "git gud levels --skills" to print all the skills.')  # noqa: E501
                     return
             elif self.is_initialized():
-                current_skill = self.get_level().skill
+                current_skill = get_operator().get_level().skill
                 skills_to_show = [current_skill]
                 print('Levels in the current skill "{}" :'.format(current_skill.name))  # noqa: E501
             else:
@@ -417,71 +377,90 @@ class GitGud:
         print()
         show_skill_tree(
             skills_to_show,
+            bool(get_operator()),
             expand_skills=True,
             show_human_names=not args.opt_short
         )
         print()
         print("Load a level with `git gud load`")
 
-    def handle_load(self, args):
-        self.assert_initialized(skip_level_check=True)
+    def load_level_by_direction(self, load_direction, force):
+        if load_direction == "prev":
+            load_direction = "previous"
 
-        argskillset = args.skill_name.split("-", 1)
+        try:
+            level = get_operator().get_level()
+        except InitializationError as e:
+            print(e)
+            print(f"Cannot load {load_direction} level.")
+            return
 
-        # Set up args.level_name and args.skill_name
-        if args.level_name:
-            if args.skill_name == "-":
-                # Replace the dash with the current skill's name.
-                args.skill_name = self.get_level().skill.name
-        else:
-            if len(argskillset) == 2:
-                args.skill_name, args.level_name = tuple(argskillset)
+        if load_direction == "next":
+            if level.has_ever_been_completed() or force:
+                level_to_load = level.next_level
             else:
-                args.skill_name, args.level_name = argskillset[0], None
-
-        skill_to_load = self.get_level().skill.name
-        if args.skill_name:
-            if args.skill_name.lower() in {"next", "prev", "previous"}:
-                query = args.skill_name.lower()
-                level = self.get_level()
-
-                if query == "next":
-                    level_to_load = level.next_level
-                else:
-                    query = "previous"
-                    level_to_load = level.prev_level
-
-                if level_to_load is not None:
-                    self.load_level(level_to_load)
-                else:
-                    if query == "next":
-                        all_levels_complete()
-                    else:
-                        print('Already on the first level. To reload the level, use "git gud reload".')  # noqa: E501
-                    print('\nTo view levels/skills, use "git gud levels --all"')  # noqa: E501
+                handle_load_confirm()
                 return
-            else:
-                skill_to_load = args.skill_name
-
-        level_to_load = '1'
-        if args.level_name:
-            level_to_load = args.level_name
-
-        if skill_to_load in all_skills.keys():
-            skill = all_skills[skill_to_load]
-            if level_to_load in skill.keys():
-                level = skill[level_to_load]
-                self.load_level(level)
-            else:
-                print('Level "{}" does not exist'.format(args.level_name))
-                print('\nTo view levels/skills, use "git gud levels --all"')  # noqa: E501
         else:
-            print('Skill "{}" does not exist'.format(args.skill_name))
+            level_to_load = level.prev_level
+
+        if level_to_load is not None:
+            self.load_level(level_to_load)
+        else:
+            if load_direction == "next":
+                all_levels_complete()
+            else:
+                print('Already on the first level. To reload the level, use "git gud reload".')  # noqa: E501
             print('\nTo view levels/skills, use "git gud levels --all"')  # noqa: E501
 
+    def handle_load(self, args):
+        self.assert_initialized()
+        args.skill_name = args.skill_name.lower()
+        if args.level_name:
+            args.level_name = args.level_name.lower()
+
+        if args.skill_name in {"next", "prev", "previous"}:
+            self.load_level_by_direction(args.skill_name, args.force)
+            return
+
+        # No matter what, the dash separates the skill and level
+        if '-' in args.skill_name:
+            identifier = args.skill_name + (args.level_name or '')
+        else:
+            identifier = args.skill_name + '-' + (args.level_name or '')
+        if identifier.count('-') != 1:
+            print("Load formula must not contain more than one dash.")
+            return
+
+        args.skill_name, args.level_name = identifier.split('-')
+
+        if not args.skill_name:
+            args.skill_name, loaded_level_name = get_operator() \
+                    .get_level_identifier()
+            print("Inferring skill from currently loaded level: {} {}"
+                  .format(args.skill_name, loaded_level_name))
+        if not args.level_name:
+            args.level_name = '1'
+
+        # Skill doesn't exist
+        if args.skill_name not in all_skills.keys():
+            print('Skill "{}" does not exist'.format(args.skill_name))
+            print('\nTo view levels/skills, use "git gud levels --all"')
+            return
+
+        # Level doesn't exist
+        skill = all_skills[args.skill_name]
+        if args.level_name not in skill.keys():
+            print('Level "{}" does not exist.'.format(args.level_name))
+            print('\nTo view levels/skills, use "git gud levels --all"')
+            return
+
+        level = skill[args.level_name]
+        self.load_level(level)
+
     def handle_commit(self, args):
         self.assert_initialized()
-        file_operator = operations.get_operator()
+        file_operator = get_operator()
         last_commit = file_operator.get_last_commit()
         commit_name = str(int(last_commit) + 1)
 
@@ -492,12 +471,7 @@ class GitGud:
             except ValueError:
                 pass
 
-        print_info('Created file "{}"'.format(commit_name))
-        mock_simulate('git add {}'.format(commit_name))
-        mock_simulate('git commit -m "{}"'.format(commit_name))
-
-        commit = file_operator.add_and_commit(commit_name)
-        print_info("New Commit: {}".format(commit.hexsha[:7]))
+        commit = file_operator.add_and_commit(commit_name, silent=False)
         file_operator.track_commit(commit_name, commit.hexsha)
 
         # Next "git gud commit" name
@@ -522,10 +496,25 @@ class GitGud:
         issues_website = "https://github.com/benthayer/git-gud/issues"
         webbrowser.open_new(issues_website)
 
+    def handle_debug(self, args):
+        import readline  # noqa: F401
+        import code
+        variables = globals()
+        shell = code.InteractiveConsole(variables)
+        shell.interact(
+            banner="\n".join([
+                "You are now in the Python interpreter invoked by `git gud debug`.",  # noqa: E501
+                "Your current path is " + str(Path.cwd()),
+                "To exit, type exit()"
+            ])
+        )
+
     def parse(self):
-        args, _ = self.parser.parse_known_args()
+        if len(sys.argv) >= 2 and sys.argv[1] in self.aliases:
+            sys.argv[1] = self.aliases[sys.argv[1]]
+        args, _ = self.parser.parse_known_args(sys.argv[1:])
         if args.command is None:
-            if operations.get_operator() is None:
+            if get_operator() is None:
                 print('Currently in an uninitialized directory.')
                 print('Get started by running "git gud init" in an empty directory!')  # noqa: E501
             else:
