@@ -1,6 +1,7 @@
 
 import subprocess
 
+from gitgud import operations
 from gitgud.skills import all_skills
 
 
@@ -10,11 +11,12 @@ def test_load(gg):
         ('git gud load rampup', all_skills["rampup"]["1"]),
         ('git gud load 2 detaching', all_skills["2"]["detaching"]),
         ('git gud load rampup 4', all_skills["rampup"]["4"]),
-        ('git gud load 4-octopus', all_skills["4"]["octopus"]),
+        ('git gud load 5-octopus', all_skills["5"]["octopus"]),
         ('git gud load rampup-4', all_skills["rampup"]["4"]),
         ('git gud load -2', all_skills["rampup"]["2"])
     ]
 
     for command, level in load_tests:
         subprocess.call(command, shell=True)
-        assert level == gg.get_level()
+        op = operations.get_operator()
+        assert level == op.get_level()
